import os
import sys
import numpy as np

from PyQt6.QtCore import Qt, QUrl, QSize, QTimer, QRect, QPointF
from PyQt6.QtGui import QAction, QPainter, QColor, QPen, QKeySequence
from PyQt6.QtMultimedia import QMediaPlayer, QAudioOutput, QMediaMetaData
from PyQt6.QtMultimediaWidgets import QVideoWidget
from PyQt6.QtWidgets import (
    QMainWindow,
    QApplication,
    QFileDialog,
    QHBoxLayout,
    QPushButton,
    QSlider,
    QStyle,
    QVBoxLayout,
    QWidget,
    QLabel,
)


class Player(QMainWindow):
    """A simple Media Player using Qt"""

    def __init__(self, master=None):
        QMainWindow.__init__(self, master)
        self.setWindowTitle("Media Player")

        self.mediaplayer = QMediaPlayer()

        self.media = None
        self.mfps = None

        self.create_ui()
        self.is_paused = False

    def create_ui(self):
        """Set up the user interface, signals & slots"""
        self.widget = QWidget(self)
        self.setCentralWidget(self.widget)

        btn_size = QSize(16, 16)

        # Create the video widget
        self.videoframe = QVideoWidget()

        # Create the time box
        self.htimebox = QHBoxLayout()

        # Create the time label
        self.timeLabel = QLabel()
        self.timeLabel.setText("00:00:00.000")
        self.timeLabel.setFixedHeight(24)
        self.htimebox.addWidget(self.timeLabel)

        # Create the position slider
        self.positionslider = QSlider(Qt.Orientation.Horizontal, self)
        self.positionslider.setToolTip("Position")
        self.positionslider.setRange(0, 0)
        self.positionslider.sliderMoved.connect(self.set_position)
        # Add the position slider to the time box
        self.htimebox.addWidget(self.positionslider)

        # Create the duration time box
        self.durationLabel = QLabel()
        self.durationLabel.setText("00:00:00.000")
        self.durationLabel.setFixedHeight(24)
        self.htimebox.addWidget(self.durationLabel)

        # Create the button layout
        self.hbuttonbox = QHBoxLayout()

        # Create the previous frame button
        self.previousframe = QPushButton()
        self.previousframe.setEnabled(False)
        self.previousframe.setFixedHeight(24)
        self.previousframe.setIconSize(btn_size)
        self.previousframe.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaSkipBackward)
        )
        self.previousframe.setToolTip("Previous Frame")
        self.previousframe.clicked.connect(self.on_previous_frame)
        # Add the previous frame button to the button layout
        self.hbuttonbox.addWidget(self.previousframe)

        # Create the play/pause button
        self.playbutton = QPushButton()
        self.playbutton.setEnabled(False)
        self.playbutton.setFixedHeight(24)
        self.playbutton.setIconSize(btn_size)
        self.playbutton.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        )
        self.playbutton.setToolTip("Play")
        self.playbutton.clicked.connect(self.play_pause)
        # Add the play/pause button to the button layout
        self.hbuttonbox.addWidget(self.playbutton)

        # Create the stop button
        self.stopbutton = QPushButton()
        self.stopbutton.setEnabled(False)
        self.stopbutton.setFixedHeight(24)
        self.stopbutton.setIconSize(btn_size)
        self.stopbutton.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaStop)
        )
        self.stopbutton.setToolTip("Stop")
        self.hbuttonbox.addWidget(self.stopbutton)
        # Add the stop button to the button layout
        self.stopbutton.clicked.connect(self.stop)

        # Create the next frame button
        self.nextframe = QPushButton()
        self.nextframe.setEnabled(False)
        self.nextframe.setFixedHeight(24)
        self.nextframe.setIconSize(btn_size)
        self.nextframe.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaSkipForward)
        )
        self.nextframe.setToolTip("Next Frame")
        self.nextframe.clicked.connect(self.on_next_frame)
        # Add the next frame button to the button layout
        self.hbuttonbox.addWidget(self.nextframe)
        self.hbuttonbox.addStretch(1)

        # Create the volume slider
        self.volumeslider = QSlider(Qt.Orientation.Horizontal, self)
        self.volumeslider.setMaximum(100)
        self.volumeslider.setValue(100)
        self.volumeslider.setToolTip("Volume")
        # Add the volume slider to the button layout
        self.hbuttonbox.addWidget(self.volumeslider)
        self.volumeslider.valueChanged.connect(self.set_volume)

        # Create the main layout and add the button layout and video widget
        self.vboxlayout = QVBoxLayout()
        self.vboxlayout.addWidget(self.videoframe)
        self.vboxlayout.addLayout(self.htimebox)
        self.vboxlayout.addLayout(self.hbuttonbox)

        # Add TimeLineWidget to
        self.timeline = TimeLineWidget(self)
        self.vboxlayout.addWidget(self.timeline)

        # Setup the media player
        self.mediaplayer.setVideoOutput(self.videoframe)
        self.mediaplayer.playbackStateChanged.connect(self.playback_state_changed)
        self.mediaplayer.mediaStatusChanged.connect(self.media_status_changed)
        self.mediaplayer.positionChanged.connect(self.position_changed)
        self.mediaplayer.durationChanged.connect(self.duration_changed)

        # Setup the audio output
        self.audiooutput = QAudioOutput()
        self.mediaplayer.setAudioOutput(self.audiooutput)

        # Prevent the individual UIs from getting the focus
        for ui in [
            self.playbutton,
            self.stopbutton,
            self.nextframe,
            self.previousframe,
            self.volumeslider,
            self.positionslider,
            self.videoframe,
        ]:
            ui.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Add the main layout to the main window
        self.widget.setLayout(self.vboxlayout)

        # Create menu bar
        menu_bar = self.menuBar()

        # File menu
        file_menu = menu_bar.addMenu("&File")

        # Add actions to file menu
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon)
        open_action = QAction(
            icon,
            "&Open...",
            self,
            shortcut=QKeySequence.StandardKey.Open,
            triggered=self.open_file,
        )
        close_action = QAction(
            "Quit",
            self,
            shortcut=QKeySequence.StandardKey.Close,
            triggered=self.close,
        )

        file_menu.addAction(open_action)
        file_menu.addAction(close_action)

        # Add actions to play menu
        play_menu = menu_bar.addMenu("&Play")
        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        self.play_action = QAction(
            icon,
            "Play/Pause",
            self,
            shortcut=Qt.Key.Key_Space,
            triggered=self.play_pause,
        )

        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_MediaStop)
        self.stop_action = QAction(
            icon,
            "Stop",
            self,
            shortcut=Qt.Key.Key_S,
            triggered=self.stop
        )

        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_MediaSkipBackward)
        self.previous_frame_action = QAction(
            icon,
            "Go to the previous frame",
            self,
            shortcut=Qt.Key.Key_Left,
            triggered=self.on_previous_frame,
        )

        icon = self.style().standardIcon(QStyle.StandardPixmap.SP_MediaSkipForward)
        self.next_frame_action = QAction(
            icon,
            "Go to the next frame",
            self,
            shortcut=Qt.Key.Key_Right,
            triggered=self.on_next_frame,
        )

        play_menu.addAction(self.play_action)
        play_menu.addAction(self.stop_action)
        play_menu.addAction(self.previous_frame_action)
        play_menu.addAction(self.next_frame_action)

        # Add actions to view menu
        view_menu = menu_bar.addMenu("&View")
        fullscreen_action = QAction(
            "Toggle Fullscreen",
            self,
            shortcut=Qt.Key.Key_F,
            triggered=self.on_fullscreen,
        )

        view_menu.addAction(fullscreen_action)

        self.timer = QTimer(self)
        self.timer.setInterval(100)
        self.timer.timeout.connect(self.update_time_label)

    def on_fullscreen(self):
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def play_pause(self):
        """Toggle play/pause status"""
        if self.mediaplayer.playbackState() == QMediaPlayer.PlaybackState.PlayingState:
            self.mediaplayer.pause()
        else:
            self.mediaplayer.play()

    def stop(self):
        """Stop player"""
        self.mediaplayer.stop()
        self.playbutton.setIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
        )

    def open_file(self):
        """Open a media file in a MediaPlayer"""

        dialog_txt = "Open Video File"
        filename, _ = QFileDialog.getOpenFileName(
            self, dialog_txt, os.path.expanduser("~"), "Video Files (*.mp4)"
        )
        if not filename:
            return

        # getOpenFileName returns a tuple, so use only the actual file name
        self.media = QUrl.fromLocalFile(filename)

        # Enable the buttons
        self.playbutton.setEnabled(True)
        self.stopbutton.setEnabled(True)

        # Put the media in the media player
        self.mediaplayer.setSource(self.media)

        # Set the title of the track as window title
        self.setWindowTitle(filename)

        # Show first frame
        self.mediaplayer.play()
        self.mediaplayer.pause()

    def set_volume(self, volume):
        """Set the volume"""
        self.audiooutput.setVolume(self.volumeslider.value() / 100)

    def playback_state_changed(self, state):
        """Set the button icon when media changes state"""
        if state == QMediaPlayer.PlaybackState.PlayingState:
            self.timer.start()
            self.playbutton.setIcon(
                self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPause)
            )
        else:
            self.timer.stop()
            self.playbutton.setIcon(
                self.style().standardIcon(QStyle.StandardPixmap.SP_MediaPlay)
            )

        self.stopbutton.setEnabled(state != QMediaPlayer.PlaybackState.StoppedState)
        self.previousframe.setEnabled(state == QMediaPlayer.PlaybackState.PausedState)
        self.nextframe.setEnabled(state == QMediaPlayer.PlaybackState.PausedState)

    def media_status_changed(self, state):
        if state == QMediaPlayer.MediaStatus.LoadedMedia:
            # Check if metadata is available
            metadata = self.mediaplayer.metaData()
            if metadata:
                # If metadata is available, set the frame rate
                fps = metadata.value(QMediaMetaData.Key.VideoFrameRate)
                self.mfps = int(1000 / fps)
            else:
                self.mfps = None

    def position_changed(self, position):
        """Update the position slider"""
        self.positionslider.setValue(position)
        self.timeLabel.setText(self.milliseconds_to_formatted_string(position))
        self.timeline.set_value(position)
        self.timeline.update()

    def duration_changed(self, duration):
        """Update the duration slider"""
        self.positionslider.setRange(0, duration)
        self.durationLabel.setText(self.milliseconds_to_formatted_string(duration))
        self.timeline.set_max_value(duration)
        self.timeline.update()

    def set_position(self, position):
        """Set the position"""
        self.mediaplayer.setPosition(position)

    def on_previous_frame(self):
        """Set the position to the previous frame"""
        state = self.mediaplayer.playbackState()
        if self.mfps is None or state != QMediaPlayer.PlaybackState.PausedState:
            return
        self.set_position(self.mediaplayer.position() - self.mfps)

    def on_next_frame(self):
        """Set the position to the next frame"""
        state = self.mediaplayer.playbackState()
        if self.mfps is None or state != QMediaPlayer.PlaybackState.PausedState:
            return
        self.set_position(self.mediaplayer.position() + self.mfps)

    def update_time_label(self):
        """Update the time label"""
        self.timeLabel.setText(
            self.milliseconds_to_formatted_string(self.mediaplayer.position())
        )

    def milliseconds_to_formatted_string(self, milliseconds):
        """
        Converts milliseconds to a string in the format hh:mm:ss.ssss.
        """

        # Convert milliseconds to seconds
        total_seconds = milliseconds / 1000

        # Extract hours, minutes, seconds
        hours, remainder = divmod(total_seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        # Format time string with leading zeros
        time_string = f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"

        # Extract milliseconds (avoiding floating-point rounding issues)
        milliseconds = milliseconds % 1000
        millisecond_string = f"{milliseconds:03d}"  # Pad with leading zeros

        return f"{time_string}.{millisecond_string}"


class TimeLineWidget(QWidget):

    def __init__(self, player=None):
        """Initializes the timeline widget"""
        super().__init__()
        self.player = player
        self.setMinimumSize(1, 60)
        self.setFixedHeight(60)
        self.value = 0
        self.min_value = 0
        self.max_value = 1
        self.main_rect = None

    def set_value(self, value):
        """Sets the value opf the timeline"""
        self.value = value

    def set_min_value(self, value):
        """Sets the min"""
        self.min_value = value

    def set_max_value(self, value):
        """Sets the max"""
        self.max_value = value

    def paintEvent(self, e):
        """Handles the painting of the widget"""
        qp = QPainter()
        qp.begin(self)
        self.draw_widget(qp)
        qp.end()

    def draw_widget(self, qp):
        """Draws the widget on the painter"""
        size = self.size()
        w = size.width()
        h = size.height()
        y = 30
        line_height = 5
        duration_in_seconds = self.max_value / 1000
        step = w / duration_in_seconds

        # Define the main rectangle
        self.main_rect = QRect(0, y, w - 1, h - y - 1)

        # Calcul the size of the rectangle based on time passed
        till = int(((w / self.max_value) * self.value))

        # Draw the rectangle
        qp.setPen(QColor(255, 255, 255))
        qp.setBrush(QColor(209, 209, 48))
        qp.drawRect(0, y, till, self.main_rect.height())

        # Draw the main rectangle
        pen = QPen(QColor(20, 20, 20), 1, Qt.PenStyle.SolidLine)
        qp.setPen(pen)
        qp.setBrush(Qt.BrushStyle.NoBrush)
        qp.drawRect(self.main_rect)

        j = 1
        for i in np.arange(step, self.max_value, step):
            # Calcul the position of the line
            qp.drawLine(QPointF(i, y), QPointF(i, y - line_height))

            # Calcul the size of the text
            metrics = qp.fontMetrics()
            fw = metrics.horizontalAdvance(str(j))

            # Calcul the position of the text
            posText = QPointF(i - fw / 2, (h - y + line_height) / 2)

            # Draw the text
            qp.drawText(posText, str(j))
            j += 1

    def mousePressEvent(self, event):
        """Handles mouse press events"""
        if event.button() == Qt.MouseButton.LeftButton:
            new_value = int(event.position().x() * self.max_value / self.size().width())
            self.player.set_position(new_value)

    def mouseMoveEvent(self, event):
        """Handles mouse movement events"""
        new_value = int(event.position().x() * self.max_value / self.size().width())
        self.player.set_position(new_value)


def main():
    """Entry point for our simple vlc player"""
    app = QApplication(sys.argv)
    player = Player()
    player.show()
    player.resize(640, 480)
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
